## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

from .version import VERSION as __version__  # noqa

__author__ = "Kaden Bilyeu (Bikatr7) <Bikatr7@proton.me>"

from .elucidate import Elucidate

__all__ = ["Elucidate"]


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in imports
import typing
import asyncio

## custom modules
from ..protocols.openai_service_protocol import OpenAIServiceProtocol

from ..util.classes import SystemTranslationMessage, ModelTranslationMessage, ChatCompletion, NOT_GIVEN, openai_service
from ..util.attributes import VALID_JSON_OPENAI_MODELS, _sync_logging_decorator, _async_logging_decorator

##-------------------start-of-attributes---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

_default_evaluation_instructions = SystemTranslationMessage("Please suggest a revised of the given text given it's original text and it's translation.")

##-------------------start-of-_build_evaluation_batches()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
def _build_evaluation_batches(text: typing.Union[str, typing.Iterable[str], ModelTranslationMessage, typing.Iterable[ModelTranslationMessage]],
                            instructions: typing.Optional[typing.Union[str, SystemTranslationMessage]] = None) -> typing.List[typing.Tuple[ModelTranslationMessage, SystemTranslationMessage]]:
    
    """
    
    Builds the evaluation batches for the OpenAI service.

    Parameters:
    text (string | iterable[string] | ModelTranslationMessage | iterable[ModelTranslationMessage]) : The text to evaluate.
    instructions (string | SystemTranslationMessage) : The instructions to use for the evaluation.

    Returns:
    evaluation_batches (list[tuple[ModelTranslationMessage, SystemTranslationMessage]]) : The evaluation batches.

    """

    
    if(isinstance(instructions, str)):
        instructions = SystemTranslationMessage(instructions)

    elif(not isinstance(instructions, SystemTranslationMessage)):
        raise ValueError("Invalid type for instructions. Must either be a string or a pre-built SystemTranslationMessage object.")

    if(isinstance(text, str)):
        text = [ModelTranslationMessage(content=text)]

    elif(isinstance(text, ModelTranslationMessage)):
        text = [text]
    
    elif(isinstance(text, typing.Iterable)):
        text = [ModelTranslationMessage(content=item) if isinstance(item, str) else item for item in text]
    else:
        raise ValueError("Invalid type for text. Must either be a string, ModelTranslationMessage, or an iterable of strings/ModelTranslationMessage.")
    
    if(any(not isinstance(item, ModelTranslationMessage) for item in text)):
        raise ValueError("Invalid type in iterable. Must be either strings or ModelTranslationMessage objects.")
    
    return [(item, instructions) for item in text]

##-------------------start-of-_evaluate_translation()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
@_sync_logging_decorator
def _evaluate_translation(evaluation_instructions:typing.Optional[SystemTranslationMessage],
                            evaluation_prompt:ModelTranslationMessage,
                            service:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                            ) -> ChatCompletion:
    
    """
    
    Synchronously evaluates the text using the OpenAI API.

    Parameters:
    evaluation_instructions (SystemTranslationMessage) : The instructions to use for the evaluation.
    evaluation_prompt (ModelTranslationMessage) : The text to evaluate.

    Returns:
    response (ChatCompletion) : The response from the API.

    """
    
    if(evaluation_instructions is None):
        evaluation_instructions = service._default_evaluation_instructions

    if(service._decorator_to_use is None):
        return service.__evaluate_translation(evaluation_instructions, evaluation_prompt)

    decorated_function = service._decorator_to_use(service.__evaluate_translation)
    return decorated_function(evaluation_instructions, evaluation_prompt)

##-------------------start-of-_evaluate_translation_async()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
@_async_logging_decorator
async def _evaluate_translation_async(evaluation_instructions:typing.Optional[SystemTranslationMessage],
                            evaluation_prompt:ModelTranslationMessage,
                            service:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                            ) -> ChatCompletion:
    
    """

    Asynchronously evaluates the text using the OpenAI API.
    
    Parameters:
    evaluation_instructions (SystemTranslationMessage) : The instructions to use for the evaluation.
    evaluation_prompt (ModelTranslationMessage) : The text to evaluate.

    Returns:
    response (ChatCompletion) : The response from the API.

    """
    
    if(evaluation_instructions is None):
        evaluation_instructions = service._default_translation_instructions

    if(service._decorator_to_use is None):
        return await service.__evaluate_translation_async(evaluation_instructions, evaluation_prompt)
    
    decorated_function = service._decorator_to_use(service.__evaluate_translation_async)
    return await decorated_function(evaluation_instructions, evaluation_prompt)

##-------------------start-of-internal_evaluate_translation()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
def internal_evaluate_translation(instructions:SystemTranslationMessage, 
                            prompt:ModelTranslationMessage,
                            service:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                            ) -> ChatCompletion:

    """

    Synchronously evaluates the text using the OpenAI API.

    Parameters:
    instructions (SystemTranslationMessage) : The instructions to use for the evaluation.
    prompt (ModelTranslationMessage) : The text to evaluate.

    Returns:
    response (ChatCompletion) : The response from the API.

    """

    response_format = "json_object" if service._json_mode and service._model in VALID_JSON_OPENAI_MODELS else "text"

    attributes = ["temperature", "logit_bias", "top_p", "n", "stream", "stop", "presence_penalty", "frequency_penalty", "max_tokens"]
    message_args = {
        "response_format": { "type": response_format },
        "model": service._model,
        "messages": [instructions.to_dict(), prompt.to_dict()],
        ## scary looking dict comprehension to get the attributes that are not NOT_GIVEN
        **{attr: getattr(service, f"_{attr}") for attr in attributes if getattr(service, f"_{attr}") != NOT_GIVEN}
    }

    response = service._sync_client.chat.completions.create(**message_args)
    
    return response

##-------------------start-of-internal_evaluate_translation_async()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
async def internal_evaluate_translation_async(instructions:SystemTranslationMessage, 
                                 prompt:ModelTranslationMessage,
                                 service:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                                 ) -> ChatCompletion:

    """

    Asynchronously evaluates the text using the OpenAI API.

    Parameters:
    instructions (SystemTranslationMessage) : The instructions to use for the evaluation.
    prompt (ModelTranslationMessage) : The text to evaluate.

    Returns:
    response (ChatCompletion) : The response from the API.

    """

    async with service._semaphore:

        response_format = "json_object" if service._json_mode and service._model in VALID_JSON_OPENAI_MODELS else "text"

        if(service._rate_limit_delay is not None):
            await asyncio.sleep(service._rate_limit_delay)

        attributes = ["temperature", "logit_bias", "top_p", "n", "stream", "stop", "presence_penalty", "frequency_penalty", "max_tokens"]
        message_args = {
            "response_format": { "type": response_format },
            "model": service._model,
            "messages": [instructions.to_dict(), prompt.to_dict()],
            ## scary looking dict comprehension to get the attributes that are not NOT_GIVEN
            **{attr: getattr(service, f"_{attr}") for attr in attributes if getattr(service, f"_{attr}") != NOT_GIVEN}
        }

        response = await service._async_client.chat.completions.create(**message_args)
        
        return response


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## custom modules 
from .util.classes import openai_service

from .evaluators.openai_evaluator import _default_evaluation_instructions, _evaluate_translation, internal_evaluate_translation, _build_evaluation_batches, _evaluate_translation_async, internal_evaluate_translation_async

def monkeystrap():

    ## monkeystrapping new sync functions to OpenAIService
    setattr(openai_service.OpenAIService, "_evaluate_translation", _evaluate_translation)
    setattr(openai_service.OpenAIService, "__evaluate_translation", internal_evaluate_translation)
    
    setattr(openai_service.OpenAIService, "_build_evaluation_batches", _build_evaluation_batches)

    ## monkeystrapping new async functions to OpenAIService
    setattr(openai_service.OpenAIService, "_evaluate_translation_async", _evaluate_translation_async)
    setattr(openai_service.OpenAIService, "__evaluate_translation_async", internal_evaluate_translation_async)

    ## monkeystrapping new attributes to OpenAIService
    setattr(openai_service.OpenAIService, "_default_evaluation_instructions", _default_evaluation_instructions)


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in imports
import typing
import asyncio

## third-party imports
import google.generativeai as genai

## custom modules
from ..util.classes import GenerationConfig, GenerateContentResponse

class GeminiServiceProtocol(typing.Protocol):

    _default_translation_instructions:str = "Please translate the following text into English."
    _default_model:str = "gemini-pro"

    _system_message = _default_translation_instructions

    _model:str = _default_model
    _temperature:float = 0.5
    _top_p:float = 0.9
    _top_k:int = 40
    _candidate_count:int = 1
    _stream:bool = False
    _stop_sequences:typing.List[str] | None = None
    _max_output_tokens:int | None = None

    _client:genai.GenerativeModel
    _generation_config:GenerationConfig

    _semaphore_value:int = 5
    _semaphore:asyncio.Semaphore = asyncio.Semaphore(_semaphore_value)

    _rate_limit_delay:float | None = None

    _decorator_to_use:typing.Union[typing.Callable, None] = None

    _log_directory:str | None = None

    ## Set to prevent any blockage of content
    _safety_settings = [
        {
            "category": "HARM_CATEGORY_DANGEROUS",
            "threshold": "BLOCK_NONE",
        },
        {
            "category": "HARM_CATEGORY_HARASSMENT",
            "threshold": "BLOCK_NONE",
        },
        {
            "category": "HARM_CATEGORY_HATE_SPEECH",
            "threshold": "BLOCK_NONE",
        },
        {
            "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
            "threshold": "BLOCK_NONE",
        },
        {
            "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
            "threshold": "BLOCK_NONE",
        },
    ]

    _json_mode:bool = False
    _response_schema:typing.Mapping[str, typing.Any] | None = None
    
    @staticmethod
    def _redefine_client() -> None: ...

    @staticmethod
    def _evaluate_translation(text_to_translate:str
                        ) -> GenerateContentResponse: ...
    @staticmethod
    def __evaluate_translation(text_to_translate:str
                                    ) -> GenerateContentResponse: ...
    
    @staticmethod
    def _set_attributes(model:str="gemini-pro",
                        system_message:str | None = _default_translation_instructions,
                        temperature:float=0.5,
                        top_p:float=0.9,
                        top_k:int=40,
                        candidate_count:int=1,
                        stream:bool=False,
                        stop_sequences:typing.List[str] | None=None,
                        max_output_tokens:int | None=None,
                        decorator:typing.Union[typing.Callable, None]=None,
                        logging_directory:str | None=None,
                        semaphore:int | None=None,
                        rate_limit_delay:float | None=None,
                        json_mode:bool=False,
                        response_schema:typing.Mapping[str, typing.Any] | None = None
                        ) -> None: ...


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## third-party imports
from openai import AsyncOpenAI, OpenAI

## custom modules
from .imports.easytl_importer import openai_service, gemini_service

from .imports.easytl_importer import SystemTranslationMessage, ModelTranslationMessage, ChatCompletion, NOT_GIVEN, NotGiven, GenerationConfig, GenerateContentResponse


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in libraries
import asyncio
import os
import sys
import time
import logging

## third-party libraries
import backoff

from elucidate import Elucidate

from elucidate.exceptions import OpenAIError

##-------------------start-of-read_api_key()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

def read_api_key(filename):

    try:
        with open(filename, 'r') as file:
            return file.read().strip()
    except:
        pass

##-------------------start-of-setup_preconditions()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

def setup_preconditions():

    openai_api_key = os.environ.get('OPENAI_API_KEY')

    ## the following will replace the keys with local test keys if the environment variables are not set (not a github actions environment)
    logging_directory = os.getenv('LOGGING_DIRECTORY', '/tmp/')

    if(openai_api_key is None):
        openai_api_key = read_api_key("tests/openai.txt")
        logging_directory = "tests/"

    ## if any of these trigger, something is wrong (not in local test environment or github actions environment)
    assert openai_api_key is not None, "OPENAI_API_KEY environment variable must be set"

    ## set the credentials for the services
    Elucidate.set_credentials("openai", openai_api_key)

    return logging_directory

##-------------------start-of-main()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

async def main():

    logging_directory = setup_preconditions()

    ## probably self explanatory from this point on

    decorator = backoff.on_exception(backoff.expo, exception=(OpenAIError), logger=logging.getLogger())

    print("------------------------------------------------OpenAI------------------------------------------------")

    print("-----------------------------------------------Text response-----------------------------------------------")

    print(Elucidate.openai_evaluate("山村美紀の独白\nYamamura Miki's Monologue", model="gpt-3.5-turbo", evaluation_instructions="Please suggest a revised English translation based on the original Japanese text. Do not change if the translation is already correct.", decorator=decorator))

    print(await Elucidate.openai_evaluate_async("山村美紀の独白\nYamamura Miki's Monologue", model="gpt-3.5-turbo", evaluation_instructions="Please suggest a revised English translation based on the original Japanese text. Do not change if the translation is already correct.", decorator=decorator))

##-------------------end-of-main()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

if(__name__ == "__main__"):
    ## setup logging
    logging.basicConfig(level=logging.DEBUG, 
                        filename='passing.log',
                        filemode='w', 
                        format='[%(asctime)s] [%(levelname)s] [%(filename)s] %(message)s', 
                        datefmt='%Y-%m-%d %H:%M:%S')

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] [%(filename)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)

    asyncio.run(main())



## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

from .util.imports.easytl_importer import InvalidResponseFormatException, InvalidTextInputException, EasyTLException, InvalidAPITypeException, OpenAIError

class ElucidateException(EasyTLException):
    
    def __init__(self, message:str):
        super().__init__(message)


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in imports
import typing
import asyncio

## custom modules
from ..util.classes import SystemTranslationMessage, ModelTranslationMessage, ChatCompletion, NOT_GIVEN, NotGiven, OpenAI, AsyncOpenAI
from ..util.attributes import  _sync_logging_decorator, _async_logging_decorator

class OpenAIServiceProtocol(typing.Protocol):

    _default_translation_instructions:typing.ClassVar[SystemTranslationMessage]

    _default_evaluation_instructions:typing.ClassVar[SystemTranslationMessage] = SystemTranslationMessage("Please suggest a revised of the given text given it's original text and it's translation.")
    _system_message:typing.Optional[typing.Union[SystemTranslationMessage, str]] = _default_evaluation_instructions

    _log_directory:str | None = None

    _decorator_to_use: typing.Union[typing.Callable, None]

    _json_mode:bool

    _rate_limit_delay:float | None = None

    _default_model:str = "gpt-4"
    _model:str

    _sync_client:OpenAI
    _async_client:AsyncOpenAI

    _semaphore_value:int = 5
    _semaphore:asyncio.Semaphore = asyncio.Semaphore(_semaphore_value)


    @staticmethod
    def _build_evaluation_batches(text: typing.Union[str, typing.Iterable[str], ModelTranslationMessage, typing.Iterable[ModelTranslationMessage]],
                                instructions: typing.Optional[typing.Union[str, SystemTranslationMessage]] = None) -> typing.List[typing.Tuple[ModelTranslationMessage, SystemTranslationMessage]]: ...

    @staticmethod
    @_sync_logging_decorator
    def _evaluate_translation(evaluation_instructions:typing.Optional[SystemTranslationMessage],
                                evaluation_prompt:ModelTranslationMessage,
                                ) -> ChatCompletion: ...
    @staticmethod
    @_async_logging_decorator
    async def _evaluate_translation_async(evaluation_instructions:typing.Optional[SystemTranslationMessage],
                            evaluation_prompt:ModelTranslationMessage
                            ) -> ChatCompletion: ...

    @staticmethod
    def __evaluate_translation(instructions: SystemTranslationMessage, 
                               prompt: ModelTranslationMessage) -> ChatCompletion: ...
    
    @staticmethod
    async def __evaluate_translation_async(instructions: SystemTranslationMessage,
                                    prompt: ModelTranslationMessage) -> ChatCompletion: ...
    
    @staticmethod
    def _set_attributes(model:str = _default_model,
                        temperature:float | None | NotGiven = NOT_GIVEN,
                        logit_bias:typing.Dict[str, int] | None | NotGiven = NOT_GIVEN,
                        top_p:float | None | NotGiven = NOT_GIVEN,
                        n:int | None | NotGiven = 1,
                        stream:bool = False,
                        stop:typing.List[str] | None | NotGiven = NOT_GIVEN,
                        max_tokens:int | None | NotGiven = NOT_GIVEN,
                        presence_penalty:float | None | NotGiven = NOT_GIVEN,
                        frequency_penalty:float | None | NotGiven = NOT_GIVEN,
                        decorator:typing.Union[typing.Callable, None]=None,
                        logging_directory:str | None=None,
                        semaphore:int | None=None,
                        rate_limit_delay:float | None=None,
                        json_mode:bool=False
                        ) -> None: ...



## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in libraries
import logging
import typing
import asyncio

## custom modules 
from .protocols.openai_service_protocol import OpenAIServiceProtocol

from .util.classes import openai_service

from .monkeystrapper import monkeystrap

## monkeystrapping new functions to EasyTL
monkeystrap()

## finally importing EasyTL with modified classes
from easytl import EasyTL

from .util.classes import ModelTranslationMessage, SystemTranslationMessage, ChatCompletion, NOT_GIVEN, NotGiven
from .util.attributes import _validate_easytl_llm_translation_settings, _return_curated_openai_settings, _validate_stop_sequences, _validate_text_length, _is_iterable_of_strings

from .exceptions import InvalidResponseFormatException, InvalidTextInputException, ElucidateException, InvalidAPITypeException

class Elucidate:

    """
    
    Elucidate global client.


    """
    
##-------------------start-of-openai_evaluate()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def openai_evaluate(text:typing.Union[str, typing.Iterable[str], ModelTranslationMessage, typing.Iterable[ModelTranslationMessage]],
                        override_previous_settings:bool = True,
                        decorator:typing.Callable | None = None,
                        logging_directory:str | None = None,
                        response_type:typing.Literal["text", "raw", "json", "raw_json"] | None = "text",
                        evaluation_delay:float | None = None,
                        evaluation_instructions:str | SystemTranslationMessage | None = None,
                        model:str="gpt-4",
                        temperature:float | None | NotGiven = NOT_GIVEN,
                        top_p:float | None | NotGiven = NOT_GIVEN,
                        stop:typing.List[str] | None | NotGiven = NOT_GIVEN,
                        max_tokens:int | None | NotGiven = NOT_GIVEN,
                        presence_penalty:float | None | NotGiven = NOT_GIVEN,
                        frequency_penalty:float | None | NotGiven = NOT_GIVEN,
                        _protocol:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                        ) -> typing.Union[typing.List[str], str, typing.List[ChatCompletion], ChatCompletion]:
        
        """

        Performs an evaluation on already translated text using the original untranslated text with OpenAI. Your text attribute should contain both.

        This function assumes that the API key has already been set.

        Evaluation instructions default to 'Please suggest a revised of the given text given it's original text and it's translation.' if not specified.

        Due to how OpenAI's API works, NOT_GIVEN is treated differently than None. If a parameter is set to NOT_GIVEN, it is not passed to the API. If it is set to None, it is passed to the API as None.
        
        This function is not for use for real-time evaluation, nor for generating multiple response candidates. Another function may be implemented for this given demand.

        Parameters:
        text (string or iterable) : The text to evaluate.  This should be the original untranslated text along with the translated text.
        override_previous_settings (bool) : Whether to override the previous settings that were used during the last call to an OpenAI evaluation function.
        decorator (callable or None) : The decorator to use when evaluating. Typically for exponential backoff retrying. If this is None, OpenAI will retry the request twice if it fails.
        logging_directory (string or None) : The directory to log to. If None, no logging is done. This'll append the text result and some function information to a file in the specified directory. File is created if it doesn't exist. Currently broken.
        response_type (literal["text", "raw", "json", "raw_json"]) : The type of response to return. 'text' returns the evaluated text, 'raw' returns the raw response, a ChatCompletion object, 'json' returns a json-parseable string. 'raw_json' returns the raw response, a ChatCompletion object, but with the content as a json-parseable string.
        evaluation_delay (float or None) : If text is an iterable, the delay between each evaluation. Default is none. This is more important for asynchronous evaluations where a semaphore alone may not be sufficient.
        evaluation_instructions (string or SystemTranslationMessage or None) : The evaluation instructions to use. If None, the default system message is used. If you plan on using the json response type, you must specify that you want a json output and it's format in the instructions. The default system message will ask for a generic json if the response type is json.
        model (string) : The model to use. (E.g. 'gpt-4', 'gpt-3.5-turbo-0125', 'gpt-4o', etc.)
        temperature (float) : The temperature to use. The higher the temperature, the more creative the output. Lower temperatures are typically better for translation and evaluation.
        top_p (float) : The nucleus sampling probability. The higher the value, the more words are considered for the next token. Generally, alter this or temperature, not both.
        stop (list or None) : String sequences that will cause the model to stop evaluation if encountered, generally useless.
        max_tokens (int or None) : The maximum number of tokens to output.
        presence_penalty (float) : The presence penalty to use. This penalizes the model from repeating the same content in the output.
        frequency_penalty (float) : The frequency penalty to use. This penalizes the model from using the same words too frequently in the output.

        Returns:
        result (string or list - string or ChatCompletion or list - ChatCompletion) : The evaluation result. A list of strings if the input was an iterable, a string otherwise. A list of ChatCompletion objects if the response type is 'raw' and input was an iterable, a ChatCompletion object otherwise.

        """
        
        if(logging_directory is not None):
            logging.warning("Logging is currently broken. No logs will be written.")
        
        assert response_type in ["text", "raw", "json", "raw_json"], InvalidResponseFormatException("Invalid response type specified. Must be 'text', 'raw', 'json' or 'raw_json'.")
        
        _settings = _return_curated_openai_settings(locals())

        _validate_easytl_llm_translation_settings(_settings, "openai")

        _validate_stop_sequences(stop)

        _validate_text_length(text, model, service="openai")

        ## Should be done after validating the settings to reduce cost to the user
        EasyTL.test_credentials("openai")

        json_mode = True if response_type in ["json", "raw_json"] else False
        
        if(override_previous_settings == True):
            _protocol._set_attributes(model=model,
                                        temperature=temperature,
                                        logit_bias=None,
                                        top_p=top_p,
                                        n=1,
                                        stop=stop,
                                        max_tokens=max_tokens,
                                        presence_penalty=presence_penalty,
                                        frequency_penalty=frequency_penalty,
                                        decorator=decorator,
                                        logging_directory=logging_directory,
                                        semaphore=None,
                                        rate_limit_delay=evaluation_delay,
                                        json_mode=json_mode)

            ## Done afterwards, cause default evaluation instructions can change based on set_attributes()
            evaluation_instructions = evaluation_instructions or _protocol._default_evaluation_instructions
        
        else:
            evaluation_instructions = _protocol._system_message

        assert isinstance(text, str) or _is_iterable_of_strings(text) or isinstance(text, ModelTranslationMessage) or _is_iterable_of_strings(text), InvalidTextInputException("text must be a string, an iterable of strings, a ModelTranslationMessage or an iterable of ModelTranslationMessages.")

        evaluation_batches = _protocol._build_evaluation_batches(text, evaluation_instructions)
        
        evaluations = []
        
        for _text, _evaluation_instructions in evaluation_batches:

            _result = _protocol._evaluate_translation(_evaluation_instructions, _text)

            evaluation = _result if response_type in ["raw", "raw_json"] else _result.choices[0].message.content
            
            evaluations.append(evaluation)
        
        ## If originally a single text was provided, return a single translation instead of a list
        result = evaluations if isinstance(text, typing.Iterable) and not isinstance(text, str) else evaluations[0]
        
        return result
    
##-------------------start-of-openai_evaluate_async()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    async def openai_evaluate_async(text:typing.Union[str, typing.Iterable[str], ModelTranslationMessage, typing.Iterable[ModelTranslationMessage]],
                        override_previous_settings:bool = True,
                        decorator:typing.Callable | None = None,
                        logging_directory:str | None = None,
                        response_type:typing.Literal["text", "raw", "json", "raw_json"] | None = "text",
                        semaphore:int | None = 5,
                        evaluation_delay:float | None = None,
                        evaluation_instructions:str | SystemTranslationMessage | None = None,
                        model:str="gpt-4",
                        temperature:float | None | NotGiven = NOT_GIVEN,
                        top_p:float | None | NotGiven = NOT_GIVEN,
                        stop:typing.List[str] | None | NotGiven = NOT_GIVEN,
                        max_tokens:int | None | NotGiven = NOT_GIVEN,
                        presence_penalty:float | None | NotGiven = NOT_GIVEN,
                        frequency_penalty:float | None | NotGiven = NOT_GIVEN,
                        _protocol:OpenAIServiceProtocol = typing.cast(OpenAIServiceProtocol, openai_service.OpenAIService)
                        ) -> typing.Union[typing.List[str], str, typing.List[ChatCompletion], ChatCompletion]:
        
        """

        Asynchronous version of openai_evaluate(). 
        Will generally be faster for iterables. Order is preserved.

        Performs an evaluation on already translated text using the original untranslated text with OpenAI. Your text attribute should contain both.

        This function assumes that the API key has already been set.

        Evaluation instructions default to 'Please suggest a revised of the given text given it's original text and it's translation.' if not specified.

        Due to how OpenAI's API works, NOT_GIVEN is treated differently than None. If a parameter is set to NOT_GIVEN, it is not passed to the API. If it is set to None, it is passed to the API as None.
        
        This function is not for use for real-time evaluation, nor for generating multiple response candidates. Another function may be implemented for this given demand.

        Parameters:
        text (string or iterable) : The text to evaluate.  This should be the original untranslated text along with the translated text.
        override_previous_settings (bool) : Whether to override the previous settings that were used during the last call to an OpenAI evaluation function.
        decorator (callable or None) : The decorator to use when evaluating. Typically for exponential backoff retrying. If this is None, OpenAI will retry the request twice if it fails.
        logging_directory (string or None) : The directory to log to. If None, no logging is done. This'll append the text result and some function information to a file in the specified directory. File is created if it doesn't exist. Currently broken.
        response_type (literal["text", "raw", "json", "raw_json"]) : The type of response to return. 'text' returns the evaluated text, 'raw' returns the raw response, a ChatCompletion object, 'json' returns a json-parseable string. 'raw_json' returns the raw response, a ChatCompletion object, but with the content as a json-parseable string.
        semaphore (int) : The number of concurrent requests to make. Default is 5.
        evaluation_delay (float or None) : If text is an iterable, the delay between each evaluation. Default is none. This is more important for asynchronous evaluations where a semaphore alone may not be sufficient.
        evaluation_instructions (string or SystemTranslationMessage or None) : The evaluation instructions to use. If None, the default system message is used. If you plan on using the json response type, you must specify that you want a json output and it's format in the instructions. The default system message will ask for a generic json if the response type is json.
        model (string) : The model to use. (E.g. 'gpt-4', 'gpt-3.5-turbo-0125', 'gpt-4o', etc.)
        temperature (float) : The temperature to use. The higher the temperature, the more creative the output. Lower temperatures are typically better for translation and evaluation.
        top_p (float) : The nucleus sampling probability. The higher the value, the more words are considered for the next token. Generally, alter this or temperature, not both.
        stop (list or None) : String sequences that will cause the model to stop evaluation if encountered, generally useless.
        max_tokens (int or None) : The maximum number of tokens to output.
        presence_penalty (float) : The presence penalty to use. This penalizes the model from repeating the same content in the output.
        frequency_penalty (float) : The frequency penalty to use. This penalizes the model from using the same words too frequently in the output.

        Returns:
        result (string or list - string or ChatCompletion or list - ChatCompletion) : The evaluation result. A list of strings if the input was an iterable, a string otherwise. A list of ChatCompletion objects if the response type is 'raw' and input was an iterable, a ChatCompletion object otherwise.

        """
        
        if(logging_directory is not None):
            logging.warning("Logging is currently broken. No logs will be written.")
        
        assert response_type in ["text", "raw", "json", "raw_json"], InvalidResponseFormatException("Invalid response type specified. Must be 'text', 'raw', 'json' or 'raw_json'.")
        
        _settings = _return_curated_openai_settings(locals())

        _validate_easytl_llm_translation_settings(_settings, "openai")

        _validate_stop_sequences(stop)

        _validate_text_length(text, model, service="openai")

        ## Should be done after validating the settings to reduce cost to the user
        EasyTL.test_credentials("openai")

        json_mode = True if response_type in ["json", "raw_json"] else False
        
        if(override_previous_settings == True):
            _protocol._set_attributes(model=model,
                                        temperature=temperature,
                                        logit_bias=None,
                                        top_p=top_p,
                                        n=1,
                                        stop=stop,
                                        max_tokens=max_tokens,
                                        presence_penalty=presence_penalty,
                                        frequency_penalty=frequency_penalty,
                                        decorator=decorator,
                                        logging_directory=logging_directory,
                                        semaphore=semaphore,
                                        rate_limit_delay=evaluation_delay,
                                        json_mode=json_mode)

            ## Done afterwards, cause default evaluation instructions can change based on set_attributes()
            evaluation_instructions = evaluation_instructions or _protocol._default_evaluation_instructions
        
        else:
            evaluation_instructions = _protocol._system_message

        assert isinstance(text, str) or _is_iterable_of_strings(text) or isinstance(text, ModelTranslationMessage) or _is_iterable_of_strings(text), InvalidTextInputException("text must be a string, an iterable of strings, a ModelTranslationMessage or an iterable of ModelTranslationMessages.")

        _evaluation_batches = _protocol._build_evaluation_batches(text, evaluation_instructions)

        _evaluation_tasks = []

        for _text, _evaluation_instructions in _evaluation_batches:
            _task = _protocol._evaluate_translation_async(_evaluation_instructions, _text)
            _evaluation_tasks.append(_task)

        _results = await asyncio.gather(*_evaluation_tasks)

        _results:typing.List[ChatCompletion] = _results

        assert all([hasattr(_r, "choices") for _r in _results]), ElucidateException("Malformed response received. Please try again.")

        evaluation = _results if response_type in ["raw","raw_json"] else [result.choices[0].message.content for result in _results if result.choices[0].message.content is not None]

        result = evaluation if isinstance(text, typing.Iterable) and not isinstance(text, str) else evaluation[0]

        return result
    
##-------------------start-of-evaluate()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        
    @staticmethod
    def evaluate(text:str | typing.Iterable[str] | ModelTranslationMessage | typing.Iterable[ModelTranslationMessage],
                  service:typing.Optional[typing.Literal["openai"]] = "openai",
                  **kwargs) -> typing.Union[typing.List[str], str, 
                                            typing.List[ChatCompletion], ChatCompletion]:
        
        """

        Evaluates the given text using the specified service. Your text attribute should contain both the original untranslated text and the translated text.

        Please see the documentation for the specific evaluation function for the service you want to use.

        OpenAI: openai_evaluate()

        All functions can return a list of strings or a string, depending on the input. The response type can be specified to return the raw response instead:
        OpenAI: ChatCompletion

        Parameters:
        text (str | ModelTranslationMessage | typing.Iterable[str] | typing.Iterable[ModelTranslationMessage]) : The text to evaluate. This should be the original untranslated text along with the translated text.
        service (string) : The service to use for evaluation.
        **kwargs : The keyword arguments to pass to the evaluation function.

        Returns:
        result (string or list - string or ChatCompletion or list - ChatCompletion) : The evaluation result. A list of strings if the input was an iterable, a string otherwise. A list of ChatCompletion objects if the response type is 'raw' and input was an iterable, a ChatCompletion object otherwise.
      
        """

        assert service in ["openai", "gemini", "anthropic"], InvalidAPITypeException("Invalid service specified. Must be 'openai', 'gemini' or 'anthropic'.")

        if(service == ""):
            pass

        elif(service == "openai"):
            return Elucidate.openai_evaluate(text, **kwargs)
        
##-------------------start-of-translate_async()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    
    @staticmethod
    async def evaluate_async(text:str | typing.Iterable[str],
                              service:typing.Optional[typing.Literal["openai"]] = "openai",
                              **kwargs) -> typing.Union[typing.List[str], str, 
                                                        typing.List[ChatCompletion], ChatCompletion]:

        
        """

        Asynchronous version of evaluate().
        Will generally be faster for iterables. Order is preserved.
        
        Evaluates the given text using the specified service. Your text attribute should contain both the original untranslated text and the translated text.

        Please see the documentation for the specific evaluation function for the service you want to use.

        OpenAI: openai_evaluate_async()

        All functions can return a list of strings or a string, depending on the input. The response type can be specified to return the raw response instead:
        OpenAI: ChatCompletion

        Parameters:
        text (str | ModelTranslationMessage | typing.Iterable[str] | typing.Iterable[ModelTranslationMessage]) : The text to evaluate. This should be the original untranslated text along with the translated text.
        service (string) : The service to use for evaluation.
        **kwargs : The keyword arguments to pass to the evaluation function.

        Returns:
        result (string or list - string or ChatCompletion or list - ChatCompletion) : The evaluation result. A list of strings if the input was an iterable, a string otherwise. A list of ChatCompletion objects if the response type is 'raw' and input was an iterable, a ChatCompletion object otherwise.

        """

        assert service in ["openai", "gemini", "anthropic"], InvalidAPITypeException("Invalid service specified. Must be 'openai', 'gemini' or 'anthropic'.")

        if(service == ""):
            pass

        elif(service == "openai"):
            return await Elucidate.openai_evaluate_async(text, **kwargs)

##-------------------start-of-set_credentials()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def set_credentials(api_type:typing.Literal["gemini", "openai", "anthropic"], credentials:typing.Union[str, None] = None) -> None:
        
        """

        Sets the credentials for the specified API type.

        Parameters:
        api_type (literal["gemini", "openai", "anthropic"]) : The API type to set the credentials for.
        credentials (string) : The credentials to set. This is an api key for the specified API type.

        """

        EasyTL.set_credentials(api_type, credentials)

##-------------------start-of-test_credentials()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def test_credentials(api_type:typing.Literal["gemini", "openai", "anthropic"]) -> typing.Tuple[bool, typing.Optional[Exception]]:
        
        """

        Tests the credentials for the specified API type.

        Parameters:
        api_type (literal["gemini", "openai", "anthropic"]) : The API type to test the credentials for.

        Returns:
        (bool) : Whether the credentials are valid.
        (Exception) : The exception that was raised, if any. None otherwise.

        """

        return EasyTL.test_credentials(api_type)
    
##-------------------start-of-calculate_cost()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def calculate_cost(text:str | typing.Iterable[str],
                       service:typing.Literal["gemini", "openai", "anthropic"],
                       model:typing.Optional[str] = None,
                       evaluation_instructions:typing.Optional[str] = None
                        ) -> typing.Tuple[int, float, str]:
        
        """

        Calculates the cost of evaluating the given text using the specified service.

        Parameters:
        text (string or iterable[string]) : The text to evaluate.
        service (literal["gemini", "openai", "anthropic"]) : The service to use for evaluation.
        model (string or None) : The model to use for evaluation. If None, the default model will be used.
        evaluation_instructions (string or None) : The instructions to use for evaluation. If None, the default instructions will be used.

        Returns:
        (int) : The number of tokens in the text.
        (float) : The cost of evaluating the text.
        (string) : The model used for evaluation.

        """

        return EasyTL.calculate_cost(text, service, model, evaluation_instructions)


## built-in libraries
import typing

## custom modules
from ..attributes import _convert_to_correct_type

##-------------------start-of-_return_curated_openai_settings()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

def _return_curated_openai_settings(local_settings:dict[str, typing.Any]) -> dict:
        
        """
        
        Returns the curated OpenAI settings.

        What this does is it takes local_settings from the calling function, and then returns a dictionary with the settings that are relevant to OpenAI that were converted to the correct type.

        """

        _settings = {
        "openai_model": "",
        "openai_temperature": "",
        "openai_top_p": "",
        "openai_stop": "",
        "openai_max_tokens": "",
        "openai_presence_penalty": "",
        "openai_frequency_penalty": ""
        }

        _non_openai_params = ["text", "override_previous_settings", "decorator", "translation_instructions", "logging_directory", "response_type", "response_schema", "semaphore", "translation_delay", "use_pretranslated_text"]
        _custom_validation_params = ["openai_stop"]

        for _key in _settings.keys():
            param_name = _key.replace("openai_", "")
            if(param_name in local_settings and _key not in _non_openai_params and _key not in _custom_validation_params):
                _settings[_key] = _convert_to_correct_type(_key, local_settings[param_name])

        return _settings


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## custom modules
from .imports.easytl_importer import VALID_JSON_OPENAI_MODELS,  VALID_JSON_GEMINI_MODELS, VALID_JSON_ANTHROPIC_MODELS
from .imports.easytl_importer import NOT_GIVEN
from .imports.easytl_importer import _sync_logging_decorator, _async_logging_decorator

from .imports.easytl_importer import _is_iterable_of_strings
from .imports.easytl_importer import _validate_easytl_llm_translation_settings, _return_curated_gemini_settings, _return_curated_openai_settings, _validate_stop_sequences, _validate_response_schema,  _return_curated_anthropic_settings, _validate_text_length, _convert_to_correct_type



## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## third-party imports
from easytl.services import openai_service, gemini_service

from easytl.decorators import _sync_logging_decorator, _async_logging_decorator
from easytl.classes import SystemTranslationMessage, ModelTranslationMessage, ChatCompletion, NOT_GIVEN, NotGiven, GenerationConfig, GenerateContentResponse
from easytl.util.constants import VALID_JSON_OPENAI_MODELS, VALID_JSON_ANTHROPIC_MODELS, VALID_JSON_GEMINI_MODELS
from easytl.exceptions import InvalidResponseFormatException

from easytl.util.util import _is_iterable_of_strings
from easytl.util.llm_util import _validate_easytl_llm_translation_settings, _return_curated_gemini_settings, _return_curated_openai_settings, _validate_stop_sequences, _validate_response_schema,  _return_curated_anthropic_settings, _validate_text_length, _convert_to_correct_type 

from easytl.exceptions import InvalidResponseFormatException, InvalidTextInputException, EasyTLException, InvalidAPITypeException, OpenAIError


## Copyright 2024 Kakusui LLC (https://kakusui.org) (https://github.com/Kakusui) (https://github.com/Kakusui/Elucidate)
## Use of this source code is governed by an GNU Lesser General Public License v2.1
## license that can be found in the LICENSE file.

## built-in imports
import typing
import asyncio

## third-party imports
import google.generativeai as genai

## custom modules
from ..protocols.gemini_service_protocol import GeminiServiceProtocol

from ..util.classes import SystemTranslationMessage, ModelTranslationMessage, ChatCompletion, NOT_GIVEN, gemini_service, GenerationConfig, GenerateContentResponse
from ..util.attributes import VALID_JSON_GEMINI_MODELS as VALID_SYSTEM_MESSAGE_MODELS, _sync_logging_decorator

##-------------------start-of-attributes---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

_default_evaluation_instructions = "Please suggest a revised of the given text given it's original text and it's translation."

##-------------------start-of-_redefine_client()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@staticmethod
def _redefine_client(service:GeminiServiceProtocol = typing.cast(GeminiServiceProtocol, gemini_service.GeminiService)
                        ) -> None:

    """

    Redefines the Gemini client and generation config. This should be called before making any requests to the Gemini service, or after changing any of the service's settings.

    """

    response_mime_type = "application/json" if service._json_mode else "text/plain"
    
    gen_model_params = {
        "model_name": service._model,
        "safety_settings": service._safety_settings,
        "system_instruction": service._system_message if service._model in VALID_SYSTEM_MESSAGE_MODELS else None
    }
    
    service._client = genai.GenerativeModel(**gen_model_params)
    
    generation_config_params = {
        "candidate_count": service._candidate_count,
        "stop_sequences": service._stop_sequences,
        "max_output_tokens": service._max_output_tokens,
        "temperature": service._temperature,
        "top_p": service._top_p,
        "top_k": service._top_k,
        "response_mime_type": response_mime_type,
        "response_schema": service._response_schema if service._response_schema and service._json_mode else None
    }
    
    service._generation_config = GenerationConfig(**generation_config_params)
    
    service._semaphore = asyncio.Semaphore(service._semaphore_value)

##-------------------start-of-_redefine_client_decorator()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _redefine_client_decorator(func:typing.Callable,
                                   service:GeminiServiceProtocol = typing.cast(GeminiServiceProtocol, gemini_service.GeminiService)
                                   ) -> typing.Callable:

        """

        Wraps a function to redefine the Gemini client before doing anything that requires the client.

        Parameters:
        func (callable) : The function to wrap.

        Returns:
        wrapper (callable) : The wrapped function.

        """

        def wrapper(*args, **kwargs):
            service._redefine_client() 
            return func(*args, **kwargs)
        
        return wrapper
    
##-------------------start-of-_translate_text()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    
    @staticmethod
    @_redefine_client_decorator
    @_sync_logging_decorator
    def _evaluate_translation(text_to_translate:str,
                        service:GeminiServiceProtocol = typing.cast(GeminiServiceProtocol, gemini_service.GeminiService)
                        ) -> GenerateContentResponse:

        """

        Synchronously translates text.
        Instructions default to translating whatever text is input into English.

        Parameters:
        text_to_translate (string) : The text to translate.

        Returns:
        GenerateContentResponse : The translation.

        """

        if(service._decorator_to_use is None):
            return service.__evaluate_translation(text_to_translate)

        _decorated_function = service._decorator_to_use(service.__evaluate_translation)
        return _decorated_function(text_to_translate)
    
##-------------------start-of-__evaluate_translation()---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    
    @staticmethod
    def internal_evaluate_translation(text_to_translate:str,
                                      service:GeminiServiceProtocol = typing.cast(GeminiServiceProtocol, gemini_service.GeminiService)
                                      ) -> GenerateContentResponse:

        """

        Synchronously translates text.

        Parameters:
        text_to_translate (string) : The text to translate.

        Returns:
        _response (GenerateContentResponse) : The translation.

        """

        text_request = f"{text_to_translate}" if service._model in VALID_SYSTEM_MESSAGE_MODELS else f"{service._system_message}\n{text_to_translate}"

        _response = service._client.generate_content(
            contents=text_request,
            generation_config=service._generation_config,
            safety_settings=service._safety_settings,
            stream=service._stream
        )
        
        return _response